pub struct StateWatch<K, V, T, D> {
metrics: Arc<Metrics>,
state: Arc<LockingTypedState<K, V, T, D>>,
seqno_high_water: SeqNo,
seqno_rx: Receiver<SeqNo>,
upper_rx: Receiver<Antichain<T>>,
}Expand description
A reactive subscription to changes in LockingTypedState.
This exposes two independent signals over the same state:
- Self::wait_for_seqno_ge wakes on any strict seqno advance.
- Self::wait_for_upper_past wakes only when the shard upper advances.
Invariants:
- The
state.seqnoonly advances (never regresses). This is guaranteed by LockingTypedState. seqno_high_wateris always <=state.seqno.- If
seqno_high_wateris <state.seqno, then we’ll get a notification onseqno_rx. This is maintained by notifying new seqnos under the same lock which adds them. seqno_high_wateralways holds the highest value received onseqno_rx. This is maintained by the wait method taking an exclusive reference to self.upper_rxholds the shard upper, advanced under the same lock that advances the upper. It coalesces, so a waiter always observes the latest upper.
Fields
metrics: Arc<Metrics>state: Arc<LockingTypedState<K, V, T, D>>seqno_high_water: SeqNoseqno_rx: Receiver<SeqNo>upper_rx: Receiver<Antichain<T>>Implementations
Sourceimpl<K, V, T, D> StateWatch<K, V, T, D>
impl<K, V, T, D> StateWatch<K, V, T, D>
pub(crate) fn new( state: Arc<LockingTypedState<K, V, T, D>>, metrics: Arc<Metrics>, ) -> Self
Sourcepub async fn wait_for_seqno_ge(&mut self, requested: SeqNo) -> &mut Self
pub async fn wait_for_seqno_ge(&mut self, requested: SeqNo) -> &mut Self
Blocks until the State has a SeqNo >= the requested one.
This method is cancel-safe.
Sourcepub async fn wait_for_upper_past(
&mut self,
frontier: &Antichain<T>,
) -> &mut Selfwhere
T: Timestamp,
pub async fn wait_for_upper_past(
&mut self,
frontier: &Antichain<T>,
) -> &mut Selfwhere
T: Timestamp,
Blocks until the shard upper has advanced past frontier.
Unlike Self::wait_for_seqno_ge, this only wakes on upper advances, not on seqno bumps caused by GC, rollups, since-downgrades, or no-op CaAs.
This method is cancel-safe.